{-

Block is a type class for array-like type constructors supporting features used by this project.

Note: the [] instance will be very slow, but is provided for testing and debugging.

-}

module SoundStream.Block (Block(..)) where

import qualified Data.Array.IArray as IArray
import qualified Data.Array.CArray as CArray
import Foreign.Ptr (Ptr)
import Foreign.Storable (Storable)
import qualified SoundStream.ZeroBasedCArray as ZeroBasedCArray
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)

class Block b where
  -- Conversion to and from CArrays is critical for some library functions.
  blockToCArray :: Storable e => b e -> ZeroBasedCArray e
  cArrayToBlock :: Storable e => ZeroBasedCArray e -> b e

  (!) :: Storable e => b e -> Int -> e
  a ! i = (IArray.!) (ZeroBasedCArray.toCArray (blockToCArray a)) i

  empty :: Storable e => b e
  empty = fromList []

  -- The default implementation refers to the implementation for
  -- ZeroBasedCArray, so ZeroBasedCArray's implementation must override this.
  size :: Storable e => b e -> Int
  size b = size (blockToCArray b)

  fromList :: Storable e => [e] -> b e
  fromList = cArrayToBlock . ZeroBasedCArray.fromList

  toList :: Storable e => b e -> [e]
  toList = IArray.elems . ZeroBasedCArray.toCArray . blockToCArray

  -- Like Data.Array.CArray.withCArray. In particular, the pointer should only
  -- be used during execution of the given IO action.
  withPtr :: Storable e => b e -> (Ptr e -> IO a) -> IO a
  withPtr block = CArray.withCArray (ZeroBasedCArray.toCArray (blockToCArray block))

  -- Apply a function to each element. We can't just require that b be a
  -- functor because some instances have constraints on their element types, so
  -- e.g.
  --   fmap (const "hello") x
  -- wouldn't work for Block instances that can't store strings.
  --
  -- The default implementation refers to the implementation for
  -- ZeroBasedCArray, so ZeroBasedCArray's implementation must override this.
  bmap :: (Storable e, Storable f) => (e -> f) -> b e -> b f
  bmap f = cArrayToBlock . bmap f . blockToCArray

-- The ZeroBasedCArray instance is useful because it avoids any cost converting
-- to and from CArrays.

instance Block ZeroBasedCArray where
  blockToCArray = id
  cArrayToBlock = id
  size = CArray.size . ZeroBasedCArray.toCArray
  bmap = ZeroBasedCArray.amap

-- Slow list instance. Not recommended for any real application.

instance Block [] where
  blockToCArray = ZeroBasedCArray.fromList
  cArrayToBlock = IArray.elems . ZeroBasedCArray.toCArray
  (!) = (!!)
  empty = []
  size = length
  fromList = id
  toList = id
  bmap = map
