{-

For generating sounds consisting of several tracks of pulse sounds. Each track
has an amplitude, a pitch, a duration, and a sequence of times the pulse is
played at.

Each pulse is a simple triangle-shaped envelope multiplied by a sine wave.

-}

module SoundStream.PulseTrack (PulseTrack(..), render) where

import SoundStream.Duration (Duration, toSeconds)
import SoundStream.Frequency (Frequency, toHz)
import qualified SoundStream.Sndio as Sndio
import qualified SoundStream.TimedClip as TimedClip

data PulseTrack =
  PulseTrack
  { amplitude :: Double
  , pitch :: Frequency Double
  , -- Must be nonnegative.
    duration :: Duration Double
  , -- The times at which to play the note. The note is centred on these times,
    -- i.e. it starts one half-duration before and ends one half-duration
    -- after.
    --
    -- The times must be in ascending order and the smallest must be at least
    -- duration/2.
    centerTimes :: [Duration Double]
  }

-- Renders the given PulseTracks all played at once. See the comment at the top
-- of the file. If the input is finite, the returned sound ends when no pulse
-- tracks are playing.
render :: Sndio.Parameters -> [PulseTrack] -> [Double]
render parameters tracks =
  TimedClip.render (foldr mergeSortedLists [] (map (toClips parameters) tracks))

toClips :: Sndio.Parameters -> PulseTrack -> [TimedClip.Clip Double]
toClips parameters track =
  let pulse = renderPulse parameters track
  in [ TimedClip.Clip
       { TimedClip.startingSample =
           -- We use 'ceiling' instead of 'round' to make it less likely the
           -- result will be negative.
           ceiling
           ( (toSeconds centerTime - toSeconds (duration track) / 2) *
             toHz (Sndio.sampleRate parameters)
           )
       , TimedClip.sound = pulse
       }
     | centerTime <- centerTimes track
     ]

renderPulse :: Sndio.Parameters -> PulseTrack -> [Double]
renderPulse parameters track =
  let samplesPerCycle = toHz (Sndio.sampleRate parameters) / toHz (pitch track)
      durationInSamples = toSeconds (duration track) * toHz (Sndio.sampleRate parameters)
  in [ let envelope = 2 * amplitude track * max 0 (0.5 - abs (fromIntegral t / durationInSamples - 0.5))
       in envelope * sin (2 * pi * fromIntegral t / samplesPerCycle)
     | t <- [1 .. ceiling durationInSamples - 1]
     ]

mergeSortedLists :: Ord a => [a] -> [a] -> [a]
mergeSortedLists [] l = l
mergeSortedLists l@(_:_) [] = l
mergeSortedLists l0@(h0:t0) l1@(h1:t1)
  | h0 <= h1  = h0 : mergeSortedLists t0 l1
  | otherwise = h1 : mergeSortedLists l0 t1
