{-

Slightly higher-level API on top of CSndio. Only supports audio playback (not recording).

Currently only allows one channel. If the sndio library insists on giving us
more, we emulate one channel by playing the same sound to all channels.

We request the following parameters:
* 16-bit samples (open fails if we don't get this)
* 44.1 kHz sample rate
* Underrun policy is SIO_IGNORE

Writing is always blocking (nbio_flag is false).

-}

module SoundStream.Sndio (Handle, Parameters(..), Sample, open, close, write) where

import Control.Monad (unless, when)
import Control.Monad.Except (ExceptT(ExceptT), MonadError(throwError), runExceptT)
import Control.Monad.IO.Class (MonadIO(liftIO))
import Data.Int (Int16)
import Data.List (genericReplicate)
import Foreign.C.String (withCString)
import Foreign.C.Types
import Foreign.Marshal.Alloc (allocaBytes)
import Foreign.Ptr (Ptr, castPtr, nullPtr)
import Foreign.Storable (peek, poke, sizeOf)
import GHC.IO.Exception (IOException, userError)
import SoundStream.CSndio
import SoundStream.Frequency (Frequency, fromHz)
import qualified SoundStream.Block as Block
import SoundStream.Block (Block)

data Handle = Handle { cHandle :: !(Ptr SioHdl), numChannels :: CUInt }

type Sample = Int16

data Parameters = Parameters {sampleRate :: Frequency Double}

open :: (StringError e, MonadError e m, MonadIO m) => m (Handle, Parameters)
open = do handle <- liftIO $ withCString sioDevany $ \devAny -> sioOpen devAny sioPlay nbioBlocking
          when (handle == nullPtr) $ throwStringError "sio_open failed"
          (parameters, numChannels) <- generalizeExceptT $ setParameters handle
          startResult <- liftIO $ sioStart handle
          when (startResult == 0) $ throwStringError "sio_start failed"
          return (Handle handle numChannels, parameters)

close :: MonadIO m => Handle -> m ()
close handle = liftIO $ sioClose (cHandle handle)

write :: (Block b, StringError e, MonadError e m, MonadIO m) => Handle -> b Sample -> m ()
write handle block =
  let blockForNumChannels = repeatSamples (numChannels handle) block
      numBytes = fromIntegral (Block.size blockForNumChannels * sizeOf (undefined :: Sample))
  in do numWritten <- liftIO $ Block.withPtr blockForNumChannels $ \ptr -> sioWrite (cHandle handle) (castPtr ptr) numBytes
        when (numWritten /= numBytes) $
          throwStringError ( "Wrong number of bytes written: " ++ show numBytes
                             ++ " requested but " ++ show numWritten ++ "written."
                           )

repeatSamples :: (Block b, Integral n) => n -> b Sample -> b Sample
repeatSamples n b
  | n == 1 = b
  | otherwise = Block.fromList $ concat $ map (genericReplicate n) $ Block.toList b


-- TOOD: Export StringError or use something already written.
class StringError e where
  fromString :: String -> e

instance StringError IOException where
  fromString = userError


-- Internal functions

generalizeExceptT :: (MonadError e m, MonadIO m) => ExceptT e IO a -> m a
generalizeExceptT x = do result <- liftIO $ runExceptT x
                         case result of
                           Left e -> throwError e
                           Right v -> return v

-- We don't allow fully general MonadError because it's not compatible with allocaBytes.
--
-- Returns the user-visible parameters and the number of channels.
setParameters :: StringError e => Ptr SioHdl -> ExceptT e IO (Parameters, CUInt)
setParameters handle =
  ExceptT $ allocaBytes (sizeOf (undefined :: SioPar)) $ \ parametersPtr -> runExceptT $
  do liftIO $ sioInitpar parametersPtr
     initialParameters <- liftIO $ peek parametersPtr
     let desiredParameters =
           -- When updating this, be sure to update the following:
           -- * Any code that writes sounds to sndio.
           -- * The stdout mode for test_sounds, including the usage message
           --   and writeBlockToStdout (for simplicity it's probably better if
           --   that behaviour matches this).
           initialParameters
           { bits = 16  -- bits per sample
           , bps = 2    -- bytes per sample
           , sig = 1    -- signed
           , le = sioLeNative  -- Native endianness
           , pchan = 1         -- one playback channel
           , rate = 44100      -- samples per second
           , xrun = sioIgnore
           }
     liftIO $ poke parametersPtr desiredParameters
     setparResult <- liftIO $ sioSetpar handle parametersPtr
     when (setparResult == 0) $ throwStringError "sio_setpar failed"
     getparResult <- liftIO $ sioGetpar handle parametersPtr
     when (getparResult == 0) $ throwStringError "sio_getpar failed"
     actualParameters <- liftIO $ peek parametersPtr
     unless
       ( bits actualParameters == 16 && bps actualParameters == 2
         && sig actualParameters == 1 && le actualParameters == sioLeNative
       )
       (throwStringError "Couldn't set sample format.")
     return (Parameters { sampleRate = fromHz (fromIntegral (rate actualParameters)) }, pchan actualParameters)

throwStringError :: (StringError e, MonadError e m) => String -> m a
throwStringError s = throwError (fromString s)


-- Constants

nbioBlocking :: CInt
nbioBlocking = 0
