{-

A Clip is a sound together with a time (in samples) at which it should be
played. The render function merges a sorted list of clips into one sound.

-}

module SoundStream.TimedClip (Clip(..), render) where

data Clip a =
  Clip
  { -- Must be non-negative.
    startingSample :: Int
  , sound :: [a]
  } deriving (Eq, Ord)

-- Given a list of clips sorted by starting sample, merges them together. When
-- clips overlap, their samples are added; when no clips are playing, there is
-- silence. If the input is finite, then the output will end once no more clips
-- are playing.
render :: Num a => [Clip a] -> [a]
render = renderFrom 0 []

-- Recursive helper for 'render'. @renderFrom sampleIndex mergeSoFar
-- remainingClips@ is like render, but the output starts at sample
-- @sampleIndex@, @remainingClips@ should only have 'startingSample' values at
-- least that high, and @mergeSoFar@ is added to the output.
renderFrom :: Num a => Int -> [a] -> [Clip a] -> [a]
renderFrom _ mergeSoFar [] = mergeSoFar
renderFrom sampleIndex mergeSoFar (Clip clipStart samples : clips) =
  let (samplesBeforeClipStart, samplesAfterClipStart) = splitAtPaddedWithSilence mergeSoFar (clipStart - sampleIndex)
  in samplesBeforeClipStart ++ renderFrom clipStart (mergeClips samplesAfterClipStart samples) clips

-- Like 'splitAt', but the first list always has the target length, and is
-- padded with zeroes if necessary to make that happen. The second argument
-- must be nonnegative.
splitAtPaddedWithSilence :: Num a => [a] -> Int -> ([a], [a])
splitAtPaddedWithSilence l n
  | n <= 0 = ([], l)
  | otherwise =
      case l of
        [] -> (replicate n 0, [])
        (h:t) -> let (a, b) = splitAtPaddedWithSilence t (n-1)
                 in (h:a, b)

mergeClips :: Num a => [a] -> [a] -> [a]
mergeClips [] l = l
mergeClips l@(_:_) [] = l
mergeClips (h0:t0) (h1:t1) = (h0 + h1) : mergeClips t0 t1
