module SoundStream.BlockUtil_tests (tests) where

import Data.Proxy (Proxy(Proxy), asProxyTypeOf)
import qualified SoundStream.Block as Block
import SoundStream.BlockUtil (groupListIntoBlocks)
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)
import Test.QuickCheck ((===), (==>))
import Test.Tasty (TestTree, testGroup)
import Test.Tasty.QuickCheck (testProperty)

groupListIntoBlocksTests :: TestableBlockType b => Proxy (b ()) -> TestTree
groupListIntoBlocksTests typeProxy =
  testGroup "groupListIntoBlocks"
  [ testProperty "Works with simple example" $
    groupListIntoBlocks 3 "Abracadabra" ===
    [ Block.fromList "Abr" `asBlockType` typeProxy
    , Block.fromList "aca"
    , Block.fromList "dab"
    , Block.fromList "ra"
    ]
  , testProperty "Empty gives empty" $
    \ blockSize ->
      blockSize > 0 ==>
      groupListIntoBlocks blockSize ([] :: [Char])
      === ([] `asListOfBlockType` typeProxy)
  , testProperty "All output blocks have size at most blockSize." $
    \ elements blockSize ->
      blockSize > 0 ==>
      all
      (\ block -> Block.size (block `asBlockType` typeProxy) <= blockSize)
      (groupListIntoBlocks blockSize (elements :: [Int]))
  , testProperty "All output blocks are nonempty." $
    \ elements blockSize ->
      blockSize > 0 ==>
      all
      (\ block -> Block.size (block `asBlockType` typeProxy) > 0)
      (groupListIntoBlocks blockSize (elements :: [Int]))
  , testProperty "All blocks but the last have size blockSize." $
    \ elements blockSize ->
      blockSize > 0 ==>
      let blocks = groupListIntoBlocks blockSize (elements :: [Char])
          blocksExceptLast = case reverse blocks of
            [] -> []
            (_ : t) -> t
      in all (\ block -> Block.size (block `asBlockType` typeProxy) == blockSize) blocksExceptLast
  , testProperty "Concatenating gives back original list." $
    \ elements blockSize ->
      blockSize > 0 ==>
      concat (map Block.toList (groupListIntoBlocks blockSize elements `asListOfBlockType` typeProxy))
      === (elements :: [Int])
  , testProperty "Not strict with fixed example" $
    take 2 (groupListIntoBlocks 3 ([(0 :: Int)..7] ++ error "should not be evaluated"))
    ===
    [ Block.fromList [0..2] `asBlockType` typeProxy
    , Block.fromList [3..5]
    ]
  , testProperty "Not strict in general" $
    \ elements blockSize ->
      blockSize > 0 ==>
      let numGoodBlocks = div (length elements) blockSize
          allBlocks = groupListIntoBlocks blockSize (elements ++ error "should not be evaluated")
          goodBlocks = take numGoodBlocks allBlocks `asListOfBlockType` typeProxy
      in concat (map Block.toList goodBlocks) === (take (numGoodBlocks * blockSize) elements :: [Int]) 
  ]

testsForType :: TestableBlockType b => String -> Proxy (b ()) -> TestTree
testsForType typeName typeProxy =
  testGroup ("BlockUtil with Block instance " ++ typeName)
  [ groupListIntoBlocksTests typeProxy
  ]

tests :: TestTree
tests =
  testGroup "BlockUtil"
  [ testsForType "[]" (Proxy :: Proxy [()])
  , testsForType "ZeroBasedCArray" (Proxy :: Proxy (ZeroBasedCArray ()))
  ]


-- Helpers

class (Block.Block b, Eq (b Char), Eq (b Int), Show (b Char), Show (b Int)) => TestableBlockType b where
instance TestableBlockType [] where
instance TestableBlockType ZeroBasedCArray where

asBlockType :: b c -> Proxy (b a) -> b c
asBlockType x _ = x

asListOfBlockType :: [b c] -> Proxy (b a) -> [b c]
asListOfBlockType x _ = x
