module SoundStream.Block_tests (tests) where

import qualified Control.Exception as Exception
import qualified Data.Array.IArray as IArray
import Data.Complex (Complex)
import Data.Proxy (Proxy(Proxy), asProxyTypeOf)
import qualified Foreign.Storable as Storable
import Foreign.Storable (Storable)
import qualified SoundStream.Block as Block
import SoundStream.Block ((!))
import qualified SoundStream.ZeroBasedCArray as ZeroBasedCArray
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)
import qualified Test.QuickCheck as QC
import Test.QuickCheck ((===), (==>))
import Test.Tasty (TestTree, testGroup)
import qualified Test.Tasty.HUnit as TastyHUnit
import Test.Tasty.QuickCheck (testProperty)

toFromCArrayTests :: (QC.Arbitrary e, Block.Block b, Eq e, Show e, Storable e) => String -> Proxy (b e) -> TestTree
toFromCArrayTests elementTypeName typeProxy =
  testGroup ("blockToCArray and cArrayToBlock for elements of type " ++ elementTypeName)
  [ testProperty "preserves carray contents" $
    \ elements ->
      let array = ZeroBasedCArray.fromList elements
      in Block.blockToCArray (Block.cArrayToBlock array `asProxyTypeOf` typeProxy) === array
  ]

fromToListTests :: Block.Block b => Proxy (b ()) -> TestTree
fromToListTests typeProxy =
  testGroup "fromList and toList"
  [ testProperty "fromList preserves list elements" $
    \ elements ->
      IArray.elems (ZeroBasedCArray.toCArray (Block.blockToCArray (Block.fromList elements `asProxyTypeOf` changeArgumentToInt typeProxy)))
      === elements
  , testProperty "toList . fromList is identity" $
    \ elements -> Block.toList (Block.fromList elements `asProxyTypeOf` changeArgumentToInt typeProxy) === elements
  ]

bangTests :: (QC.Arbitrary e, Block.Block b, Eq e, Show e, Storable e) => String -> Proxy (b e) -> TestTree
bangTests elementTypeName typeProxy =
  testGroup ("(!) for elements of type " ++ elementTypeName)
  [ testProperty "agrees with !! on lists" $
    \ elements i ->
      not (null elements) ==>
      let i' = mod i (length elements)
      in ((Block.fromList elements `asProxyTypeOf` typeProxy) ! i') === (elements !! i')
  ]

bangFailureTests :: Block.Block b => Proxy (b ()) -> TestTree
bangFailureTests typeProxy =
  testGroup "non-polymorphic tests"
  [ TastyHUnit.testCase "(!) rejects negative index" $
    assertError ((Block.fromList [0, 1] `asProxyTypeOf` changeArgumentToInt typeProxy) ! (-1))
  , TastyHUnit.testCase "(!) rejects too-high index" $
    assertError ((Block.fromList [0, 1] `asProxyTypeOf` changeArgumentToInt typeProxy) ! 2)
  , TastyHUnit.testCase "(!) rejects 0 index on empty array" $
    assertError ((Block.empty `asProxyTypeOf` changeArgumentToComplex typeProxy) ! 0)
  ]

emptyAndSizeTests :: Block.Block b => Proxy (b ()) -> TestTree
emptyAndSizeTests typeProxy =
  testGroup "empty and size"
  [ testProperty "empty has size zero" $
    Block.size (Block.empty `asProxyTypeOf` changeArgumentToInt typeProxy) === 0
  , testProperty "empty gives empty array" $
    let Just emptyArray = ZeroBasedCArray.fromCArray (IArray.array (0, -1) [])
    in Block.blockToCArray (Block.empty `asProxyTypeOf` changeArgumentToInt typeProxy)
       === emptyArray
  , testProperty "size measures size of list" $
    \ elements -> Block.size (Block.fromList elements `asProxyTypeOf` changeArgumentToInt typeProxy)
                  === length elements
  ]

withPtrTests :: Block.Block b => Proxy (b ()) -> TestTree
withPtrTests typeProxy =
  testGroup "withPtr"
  [ TastyHUnit.testCase "Works with empty block." $
    Block.withPtr (Block.empty `asProxyTypeOf` changeArgumentToInt typeProxy) (\_ -> return ())
  , TastyHUnit.testCase "Makes single element available." $
    do result <- Block.withPtr (Block.fromList [5, -2] `asProxyTypeOf` changeArgumentToComplex typeProxy) Storable.peek
       TastyHUnit.assertEqual "read correct value" 5 result
  , TastyHUnit.testCase "Makes other element available." $
    do result <- Block.withPtr (Block.fromList [6, 0, 15, 1] `asProxyTypeOf` changeArgumentToInt typeProxy) $
                 \ptr -> Storable.peekElemOff ptr 2
       TastyHUnit.assertEqual "read correct value" 15 result
  ]

bmapTests :: (Block.Block b, Eq (b Int), Show (b Int)) => Proxy (b ()) -> TestTree
bmapTests typeProxy =
  testGroup "bmap"
  [ testProperty "bmap agrees with map on list" $
    \ elements ->
      Block.fromList (map (+1) elements)
      ===
      (Block.bmap (+1) (Block.fromList elements) `asProxyTypeOf` changeArgumentToInt typeProxy)
  ]

-- The second argument specifies the type constructor to be tested.
testsForType :: (Block.Block b, Eq (b Int), Show (b Int)) => String -> Proxy (b ()) -> TestTree
testsForType typeName typeProxy =
  testGroup ("Block instance for " ++ typeName)
  [ toFromCArrayTests "Int" (changeArgumentToInt typeProxy)
  , toFromCArrayTests "Complex" (changeArgumentToComplex typeProxy)
  , bangTests "Int" (changeArgumentToInt typeProxy)
  , bangTests "Complex" (changeArgumentToComplex typeProxy)
  , bangFailureTests typeProxy
  , fromToListTests typeProxy
  , emptyAndSizeTests typeProxy
  , withPtrTests typeProxy
  , bmapTests typeProxy
  ]

tests :: TestTree
tests = testGroup "Block"
        [ testsForType "[]" (Proxy :: Proxy [()])
        , testsForType "ZeroBasedCArray" (Proxy :: Proxy (ZeroBasedCArray ()))
        ]

-- Helpers

assertError :: a -> IO ()
assertError x =
  do result <- Exception.try (Exception.evaluate x)
     case result of
       Left (Exception.ErrorCall _) -> return ()
       _ -> TastyHUnit.assertFailure "Expected error"

changeArgumentToComplex :: Proxy (b ()) -> Proxy (b (Complex Double))
changeArgumentToComplex = error "should not be called"

changeArgumentToInt :: Proxy (b ()) -> Proxy (b Int)
changeArgumentToInt = error "should not be called"
