module SoundStream.DFT_tests (tests) where

import Foreign.Storable (Storable)
import SoundStream.DFT
import qualified SoundStream.Block as Block
import SoundStream.Block ((!))
import qualified SoundStream.ZeroBasedCArray as ZeroBasedCArray
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)
import qualified Test.QuickCheck as QC
import Test.QuickCheck ((===), (==>))
import Test.Tasty
import Test.Tasty.QuickCheck (testProperty)

-- TODO: test instances of this class
class ApproximatelyComparable a where
  almostEqual :: a -> a -> Bool

realScalarAlmostEqual :: (Fractional a, Ord a) => a -> a -> Bool
realScalarAlmostEqual x y = abs (x - y) < 1e-4

instance ApproximatelyComparable Float where
  almostEqual = realScalarAlmostEqual

instance ApproximatelyComparable Double where
  almostEqual = realScalarAlmostEqual

instance ApproximatelyComparable e => ApproximatelyComparable [e] where
  almostEqual x y = length x == length y && and (zipWith almostEqual x y)

instance (ApproximatelyComparable e, Storable e) => ApproximatelyComparable (ZeroBasedCArray e) where
  almostEqual x y = Block.size x == Block.size y && and [(x ! i) `almostEqual` (y ! i) | i <- [0 .. Block.size x - 1]]

almostEqualProp :: (ApproximatelyComparable a, Show a) => a -> a -> QC.Property
almostEqualProp x y = QC.counterexample (show x ++ " vs. " ++ show y) $ almostEqual x y

normalizedHartleyTests :: TestTree
normalizedHartleyTests =
  testGroup "Normalized discrete Hartley transform"
  [ testProperty "Acts as identity on length-one input." $
    \ x -> normalizedHartley [x] `almostEqualProp` [x :: Double]
  , testProperty "Behaves correctly on length-two input." $
    \ x y -> normalizedHartley [x, y] `almostEqualProp` [(x + y) / sqrt 2, (x - y) / sqrt 2 :: Float]
  , testProperty "Behaves correctly on length-three input." $
    \ x y z -> normalizedHartley (Block.fromList [x, y, z] :: ZeroBasedCArray Float)
               `almostEqualProp`
               ( Block.fromList
                 [ (x + y + z) / sqrt 3
                 , (x + (-1 + sqrt 3) / 2 * y + (-1 - sqrt 3) / 2 * z) / sqrt 3
                 , (x + (-1 - sqrt 3) / 2 * y + (-1 + sqrt 3) / 2 * z) / sqrt 3
                 ]
               )
  , testProperty "Is its own inverse." $
    \ a -> not (null a) ==> normalizedHartley (normalizedHartley a :: [Double]) `almostEqualProp` a
  , testProperty "Matches formula." $
    let cas x = cos x + sin x
    in \ a -> not (null a) ==>
              let n = length a
              in normalizedHartley a
                 `almostEqualProp`
                 [ sum [ inputJ * cas (2 * pi * fromIntegral j * fromIntegral k / fromIntegral n)
                       | (j, inputJ) <- zip [0..] a
                       ]
                   / sqrt (fromIntegral n)
                   :: Double
                 | k <- [0 .. n - 1]
                 ]
  ]

tests :: TestTree
tests = testGroup "DFT" [normalizedHartleyTests]
