module SoundStream.PulseTrack_tests (tests) where

import Data.List (sort)
import SoundStream.Duration (fromSeconds, toSeconds)
import SoundStream.Frequency (fromHz, toHz)
import SoundStream.PulseTrack
import SoundStream.TestPredicates (near)
import qualified SoundStream.Sndio as Sndio
import qualified Test.QuickCheck as QC
import Test.QuickCheck ((===), (==>), (.&&.), Arbitrary(arbitrary), NonNegative(NonNegative))
import Test.Tasty (TestTree, testGroup)
import Test.Tasty.QuickCheck (testProperty)

tests :: TestTree
tests = testGroup "PulseTrack" [renderTests]

defaultParameters :: Sndio.Parameters
defaultParameters = Sndio.Parameters {Sndio.sampleRate = fromHz 1000}

renderTests :: TestTree
renderTests =
  testGroup "render"
  [ testProperty "Empty gives empty" $
    render defaultParameters [] === []
  , rendersOnePulseTests
  , testProperty "Renders two pulse tracks by adding" $
    \ amplitude0 pitch0' (SmallNonnegativeDouble duration0') (SmallNonnegativeDouble centerTime0')
      amplitude1 pitch1' (SmallNonnegativeDouble duration1') (SmallNonnegativeDouble centerTime1') ->
      let pitch0 = fromHz pitch0'
          pitch1 = fromHz pitch1'
          duration0 = fromSeconds duration0'
          duration1 = fromSeconds duration1'
          centerTime0 = fromSeconds (toSeconds duration0 / 2 + centerTime0')
          centerTime1 = fromSeconds (toSeconds duration1 / 2 + centerTime1')
          pulseTrack0 =
            PulseTrack
            { amplitude = amplitude0
            , pitch = pitch0
            , duration = duration0
            , centerTimes = [centerTime0]
            }
          pulseTrack1 =
            PulseTrack
            { amplitude = amplitude1
            , pitch = pitch1
            , duration = duration1
            , centerTimes = [centerTime1]
            }
      in render defaultParameters [pulseTrack0, pulseTrack1]
         ===
         zipWithPadded (+) (render defaultParameters [pulseTrack0]) (render defaultParameters [pulseTrack1])
  , testProperty "Renders two pulses in one track by adding" $
    \ amplitude pitch' (SmallNonnegativeDouble duration') (SmallNonnegativeDouble centerTime0') (SmallNonnegativeDouble centerTime1') ->
      let pitch = fromHz pitch'
          duration = fromSeconds (duration')
          centerTime0 = fromSeconds (toSeconds duration / 2 + centerTime0')
          centerTime1 = fromSeconds (toSeconds duration / 2 + centerTime1')
          pulseTrack0 =
            PulseTrack
            { amplitude = amplitude
            , pitch = pitch
            , duration = duration
            , centerTimes = [centerTime0]
            }
          pulseTrack1 =
            PulseTrack
            { amplitude = amplitude
            , pitch = pitch
            , duration = duration
            , centerTimes = [centerTime1]
            }
          pulseTrackTogether =
            PulseTrack
            { amplitude = amplitude
            , pitch = pitch
            , duration = duration
            , centerTimes = sort [centerTime0, centerTime1]
            }
      in render defaultParameters [pulseTrackTogether]
         ===
         zipWithPadded (+) (render defaultParameters [pulseTrack0]) (render defaultParameters [pulseTrack1])
  , testProperty "n identical tracks is the same as one multiplied by n" $
    \ n' amplitude pitch' (SmallNonnegativeDouble duration') centerTimes' ->
      let n = mod n' 15 + 1
          duration = fromSeconds duration'
          oneTrack =
            PulseTrack
            { amplitude = amplitude
            , pitch = fromHz pitch'
            , duration = duration
            , centerTimes = sort [fromSeconds (t + toSeconds duration / 2) | SmallNonnegativeDouble t <- centerTimes']
            }
          nTracks = render defaultParameters (replicate n oneTrack)
          nTimesOneTrack = map (* fromIntegral n) (render defaultParameters [oneTrack])
      in QC.counterexample (show nTracks ++ " vs. " ++ show nTimesOneTrack) $
         sum [abs (x - y) | (x, y) <- zip nTracks nTimesOneTrack] < 1e-4
  , testProperty "Respects sample rate" $
    \ (SmallNonnegativeDouble duration') (NonNegative sampleRate') (SmallNonnegativeDouble centerTime') ->
      let duration = fromSeconds duration'
          sampleRate = fromHz sampleRate'
          centerTime = fromSeconds (toSeconds duration / 2 + centerTime')
          output =
            render
            (Sndio.Parameters {Sndio.sampleRate = sampleRate})
            [ PulseTrack
              { amplitude = 1
              , pitch = fromHz 100
              , duration = duration
              , centerTimes = [centerTime]
              }
            ]
      in near 2 (length output) (round (toHz sampleRate * (toSeconds centerTime + toSeconds duration / 2)))
  , testProperty "Accepts infinitely long track" $
    \ amplitude pitch' (SmallNonnegativeDouble duration') ->
      let duration = fromSeconds duration'
          track =
            PulseTrack
            { amplitude = amplitude
            , pitch = fromHz pitch'
            , duration = duration
            , centerTimes = [fromSeconds (fromIntegral i * toSeconds duration) | i <- [1..]]
            }
      in duration > fromSeconds 1e-3 ==>
         seq (sum (take 100 (render defaultParameters [track]))) True
  ]

rendersOnePulseTests :: TestTree
rendersOnePulseTests =
  let inputAmplitude = 0.8
      output =
        render defaultParameters
        [ PulseTrack
          { amplitude = inputAmplitude
          , pitch = fromHz 300
          , duration = fromSeconds 1
          , centerTimes = [fromSeconds 1]
          }
        ]
  in testGroup "Renders one pulse" $
     [ testProperty "Duration is correct" $
       -- The sound is one second long, and is centered at time 1 second, so it
       -- ends at 1.5 seconds. With a default sample rate of 1000 Hz, that
       -- translates to 1500 samples.
       let expectedLength = 1500
           actualLength = length output
       in near 1 1500 (length output)
     , testProperty "Reaches positive amplitude" $
       near 0.1 (maximum output) inputAmplitude
     , testProperty "Reaches negative amplitude" $
       near 0.1 (minimum output) (-inputAmplitude)
     , testProperty "Confined within triangle envelope" $
       let -- In case there are off-by-one issues, we check against a slightly
           -- bigger envelope.
           relaxedEnvelope =
             replicate 499 0 ++
             [inputAmplitude * fromIntegral i / 501 | i <- [1..501]] ++
             [inputAmplitude * fromIntegral i / 501 | i <- [501,500..1]]
       in and [abs actual <= bound | (actual, bound) <- zip output relaxedEnvelope]
     ]

zipWithPadded :: (Num a, Num b) => (a -> b -> c) -> [a] -> [b] -> [c]
zipWithPadded _ [] [] = []
zipWithPadded f [] (h:t) = f 0 h : zipWithPadded f [] t
zipWithPadded f (h:t) [] = f h 0 : zipWithPadded f t []
zipWithPadded f (h0:t0) (h1:t1) = f h0 h1 : zipWithPadded f t0 t1

-- Doubles in the range 0..5 or so. Shrinking makes the number smaller.
newtype SmallNonnegativeDouble = SmallNonnegativeDouble Double

instance Arbitrary SmallNonnegativeDouble where
  arbitrary = QC.oneof [return (SmallNonnegativeDouble 0), fmap SmallNonnegativeDouble $ QC.choose (0, 5)]
  shrink (SmallNonnegativeDouble x)
    | x > 0     = [SmallNonnegativeDouble 0, SmallNonnegativeDouble (x/10)]
    | otherwise = []

instance Show SmallNonnegativeDouble where
  showsPrec prec (SmallNonnegativeDouble x) = showsPrec prec x
