module SoundStream.PureFFT_tests (tests) where

import           Data.Complex
import           SoundStream.PureFFT
import           Test.Tasty
import qualified Test.Tasty.QuickCheck as QC
import           Test.Tasty.QuickCheck ((==>))

-- TODO: test instances of this class
class ApproximatelyComparable a where
  almostEqual :: a -> a -> Bool

realScalarAlmostEqual :: (Fractional a, Ord a) => a -> a -> Bool
realScalarAlmostEqual x y = abs (x - y) < 1e-4

instance ApproximatelyComparable Float where
  almostEqual = realScalarAlmostEqual

instance ApproximatelyComparable Double where
  almostEqual = realScalarAlmostEqual

instance ApproximatelyComparable a => ApproximatelyComparable (Complex a) where
  almostEqual (xr :+ xi) (yr :+ yi) =
    almostEqual xr yr && almostEqual xi yi

instance ApproximatelyComparable a => ApproximatelyComparable [a] where
  almostEqual [] [] = True
  almostEqual [] (_ : _) = False
  almostEqual (_ : _) [] = False
  almostEqual (h0 : t0) (h1 : t1) = almostEqual h0 h1 && almostEqual t0 t1

fftTests :: TestTree
fftTests =
  testGroup "fft"
  [ QC.testProperty "length-one FFT is identity" $
    \ c -> fft [c] `almostEqual` [c :: Complex Double]
  , QC.testProperty "length-two FFT matches formula" $
    \ a b -> fft [a :: Complex Float, b] `almostEqual` [a + b, a - b]
  , QC.testProperty "length-three FFT matches formula" $
    let thirdRoot = cis (2 * pi / 3)
    in \ a b c -> fft [a :: Complex Float, b, c] `almostEqual`
                  [ a + b + c
                  , a + b * thirdRoot + c * thirdRoot^2
                  , a + b * thirdRoot^2 + c * thirdRoot
                  ]
  , QC.testProperty "length-four FFT matches formula" $
    \ a b c d -> fft [a, b, c, d :: Complex Float] `almostEqual`
                 [ a + b + c + d
                 , a + b * (0 :+ 1) - c - d * (0 :+ 1)
                 , a - b + c - d
                 , a - b * (0 :+ 1) - c + d * (0 :+ 1)
                 ]
  , QC.testProperty "FFT of one-hot is basis function" $
    \ n' i' -> let n = 1 + mod n' 100 :: Int
                   i = mod i' n
             in fft (replicate i 0 ++ (1 :: Complex Double) : replicate (n - i - 1) 0)
                `almostEqual`
                [ cis (2 * pi * fromIntegral (j * i) / fromIntegral n)
                | j <- [0 .. n - 1]
                ]
  ]

inverseFftTests :: TestTree
inverseFftTests =
  testGroup "inverseFft"
  [ QC.testProperty "fft . inverseFft is identity" $
    \ l -> not (null l) ==>
           fft (inverseFft l) `almostEqual` (l :: [Complex Double])
  , QC.testProperty "inverseFft . fft is identity" $
    \ l -> not (null l) ==>
           inverseFft (fft l) `almostEqual` (l :: [Complex Double])
  ]

tests :: TestTree
tests =
  testGroup "fft"
  [fftTests, inverseFftTests]
