-- TODO: Find a replacement for 'expectFailure' that actually does what I want.
-- For every use of 'expectFailure' in this file, I want to check that the
-- property always fails, but 'expectFailure' just checks that it fails for at
-- least some input. Asked at
-- https://github.com/nick8325/quickcheck/issues/320#issue-737265339

module SoundStream.TestPredicates_tests (tests) where

import Data.Bits.Floating (coerceToFloat)
import SoundStream.TestPredicates
import Test.QuickCheck ((==>), Arbitrary(arbitrary), Large(Large), oneof)
import Test.Tasty (TestTree, testGroup)
import Test.Tasty.QuickCheck (expectFailure, testProperty)

tests :: TestTree
tests =
  testGroup "TestPredicates"
  [ nearTests
  ]

nearTests :: TestTree
nearTests =
  testGroup "near"
  [ testProperty "Accepts exactly equal Doubles" $
    \ (DoubleWithEdgeCases x) tolerance ->
      (tolerance >= 0 && not (isNaN tolerance) && not (isNegativeZero tolerance) && not (isNaN x) && not (isInfinite x))  ==>
      near tolerance x x
  , testProperty "Accepts any non-NaN non-infinity Doubles with tolerance infinity" $
    \ (DoubleWithEdgeCases x) (DoubleWithEdgeCases y) ->
      (not (isNaN x || isInfinite x || isNaN y || isInfinite y)) ==>
      near (read "Infinity") x (y :: Double)
  , testProperty "Infinity is not close to infinity" $
    \ (DoubleWithEdgeCases tolerance) sign ->
      not (isNaN tolerance || isInfinite tolerance) && tolerance >= 0 ==>
      let signedInfinity = if sign then read "Infinity" else read "-Infinity"
      in expectFailure (near tolerance signedInfinity signedInfinity)
  , testProperty "Accepts Int exactly tolerance above" $
    \ x tolerance ->
      (tolerance >= 0 && x + tolerance >= x) ==>
      near tolerance x (x + tolerance :: Int)
  , testProperty "Accepts Int exactly tolerance below" $
    \ x tolerance ->
      (tolerance >= 0 && x - tolerance <= x) ==>
      near tolerance x (x - tolerance :: Int)
  , testProperty "Does not accept Ints tolerance + 1 above" $
    \ x tolerance ->
      (tolerance >= 0 && x + tolerance + 1 >= x) ==>
      expectFailure (near tolerance x (x + tolerance + 1 :: Int))
  , testProperty "Does not accept Ints tolerance + 1 below" $
    \ x tolerance ->
      (tolerance >= 0 && x - (tolerance + 1) <= x) ==>
      expectFailure (near tolerance x (x - (tolerance + 1) :: Int))
  , -- Double arithmetic on small integers is precise.
    testProperty "Accepts Double exactly tolerance above when arithmetic is precise" $
    \ x tolerance ->
      (tolerance >= 0 && tolerance < 10000 && abs x < 10000) ==>
      near (fromIntegral tolerance) (fromIntegral x) (fromIntegral (x - tolerance :: Int) :: Double)
  , testProperty "Accepts Double exactly tolerance below when arithmetic is precise" $
    \ x tolerance ->
      (tolerance >= 0 && tolerance < 10000 && abs x < 10000) ==>
      near (fromIntegral tolerance) (fromIntegral x) (fromIntegral (x - tolerance :: Int) :: Double)
  , testProperty "Accepts Double half tolerance above or below" $
    \ x (DoubleWithEdgeCases tolerance) sign ->
      (tolerance >= 0 && not (isNaN tolerance || isNaN x || isInfinite x)) ==>
      let offset = if sign then tolerance / 2 else - tolerance / 2
      in near tolerance x (x + offset)
  , testProperty "Does not accept unequal Doubles with zero tolerance" $
    \ (DoubleWithEdgeCases x) (DoubleWithEdgeCases y) ->
      x /= y ==>
      expectFailure (near 0 x y)
  , testProperty "Does not accept Double twice tolerance away" $
    \ tolerance x sign ->
      not (isNaN tolerance || isInfinite tolerance || isNaN x || isInfinite x) && tolerance > 0 ==>
      let offset = if sign then 2 * tolerance else -2 * tolerance
      in expectFailure (near tolerance x (x + offset :: Double))
  ]

-- Double, except the Arbitrary instance can NaN, infinity and negative zero.
-- (I don't know which of these the Double instance generates.) This generates
-- strange doubles in two ways: first, by coercing an arbitrary word to a
-- double; and second, through an incomplete list of special cases.
--
-- It could probably be included by adding more special cases; e.g. I don't
-- know how likely this is to generate a "signalling NaN", NaNs with different
-- payloads, or subnormal values.
newtype DoubleWithEdgeCases = DoubleWithEdgeCases Double

instance Arbitrary DoubleWithEdgeCases where
  arbitrary =
    let fromBits = do Large x <- arbitrary
                      return (DoubleWithEdgeCases (coerceToFloat x))
    in oneof
       ( fmap DoubleWithEdgeCases arbitrary
       : fromBits
       : map (return . DoubleWithEdgeCases . read) ["Infinity", "-Infinity", "NaN", "-0.0"]
       )

instance Show DoubleWithEdgeCases where
  showsPrec prec (DoubleWithEdgeCases x) = showsPrec prec x
