module SoundStream.TimedClip_tests (tests) where

import Data.List (sort)
import Data.Proxy (Proxy(Proxy), asProxyTypeOf)
import SoundStream.Sndio (Sample)
import SoundStream.TimedClip (Clip(Clip), render)
import qualified Test.QuickCheck as QC
import Test.QuickCheck ((===), (==>))
import Test.Tasty (TestTree, testGroup)
import Test.Tasty.QuickCheck (testProperty)

tests :: TestTree
tests = testGroup "TimedClip"
        [ renderTestsForType (Proxy :: Proxy Int)
        , renderTestsForType (Proxy :: Proxy Double)
        , renderTestsForType (Proxy :: Proxy Sample)
        ]

renderTestsForType :: (Enum a, Eq a, Num a, Ord a, QC.Arbitrary a, Show a) => Proxy a -> TestTree
renderTestsForType proxy =
  testGroup "render"
  [ testProperty "Empty gives empty" $ render [] === ([] `asProxyListTypeOf` proxy)
  , testProperty "Copies single clip" $
    \ samples -> render [Clip 0 samples] === (samples `asProxyListTypeOf` proxy)
  , testProperty "Adds delay before single clip" $
    \ delay' samples ->
      let delay = mod delay' 1000
      in render [Clip delay samples] === replicate delay 0 ++ (samples `asProxyListTypeOf` proxy)
  , testProperty "Merges samples by adding" $
    \ samplesToMerge ->
      not (null samplesToMerge) ==>
      render [Clip 0 [s] | s <- samplesToMerge] === [sum samplesToMerge `asProxyTypeOf` proxy]
  , testProperty "Merges overlapping clips" $
    render
    [ Clip 0 [1, 2, 3]
    , Clip 1 [10, 20, 30, 40]
    ]
    === [1, 12, 23, 30, 40 `asProxyTypeOf` proxy]
  , testProperty "Handles infinite input lazily" $
    take 10 (render [Clip i (repeat (1 `asProxyTypeOf` proxy)) | i <- [0..]])
    === [1..10]
  , testProperty "Output ends once all clips are done" $
    \ length0' length1' ->
      let length0 = mod length0' 1000
          length1 = mod length1' 1000
      in length (render [Clip 0 (replicate length0 1), Clip 0 (replicate length1 (1 `asProxyTypeOf` proxy))])
         === max length0 length1
  , testProperty "render's sorted-input requirement is compatible with Clip's Ord instance" $
    \ delays' ->
      not (null delays') ==>
      let delays = [mod d 1000 | d <- delays']
          sortedClips = sort [Clip delay [1 `asProxyTypeOf` proxy] | delay <- delays]
      in render sortedClips ===
         [ fromIntegral (length [() | d <- delays, d == i])
         | i <- [0..maximum delays]
         ]
  ]

asProxyListTypeOf :: [a] -> Proxy a -> [a]
asProxyListTypeOf x _ = x
