module SoundStream.ZeroBasedCArray_tests (tests) where

import qualified Data.Array as Array
import qualified Data.Array.IArray as IArray
import qualified Data.Array.CArray as CArray
import Data.Array.CArray (CArray)
import Data.Complex (Complex)
import Data.Proxy (Proxy(Proxy), asProxyTypeOf)
import Data.Maybe (isNothing)
import Foreign.Storable (Storable)
import qualified SoundStream.Block as Block
import qualified SoundStream.ZeroBasedCArray as ZeroBasedCArray
import SoundStream.ZeroBasedCArray (ZeroBasedCArray(toCArray), fromCArray)
import Test.QuickCheck ((===), (==>))
import Test.Tasty (TestTree, testGroup)
import qualified Test.Tasty.HUnit as TastyHUnit
import qualified Test.Tasty.QuickCheck as QuickCheck
import Test.Tasty.QuickCheck (testProperty)

toFromTests :: (QuickCheck.Arbitrary e, Eq e, Show e, Storable e) => Proxy e -> TestTree
toFromTests typeProxy =
  testGroup "to/from CArray"
  [ testProperty "preserves original array" $
    \ elements ->
      let array = IArray.listArray (0, length elements - 1) elements `asProxyTypeOf` (wrapProxyInCArrayInt typeProxy)
      in fmap toCArray (fromCArray array) === Just array
  , testProperty "Refuses to create non-zero-based array" $
    \ start elements ->
      start /= 0 ==>
      let array = IArray.listArray (start, start + length elements - 1) elements `asProxyTypeOf` (wrapProxyInCArrayInt typeProxy)
      in isNothing (fromCArray array)
  ]

fromListTests :: (QuickCheck.Arbitrary e, Eq e, Show e, Storable e) => Proxy e -> TestTree
fromListTests typeProxy =
  testGroup "fromList"
  [ testProperty "Preserves list elements" $
    \ elements ->
      IArray.elems (toCArray (ZeroBasedCArray.fromList elements)) === (elements `asProxyTypeOf` (wrapProxyInList typeProxy))
  ]

testsForType :: (QuickCheck.Arbitrary e, Eq e, Show e, Storable e) => String -> Proxy e -> TestTree
testsForType typeName typeProxy =
  testGroup ("for type: " ++ typeName)
  [ toFromTests typeProxy
  , fromListTests typeProxy
  ]

amapTests :: TestTree
amapTests =
  testGroup "amap"
  [ testProperty "amap agrees with map on list" $
    \ elements ->
      ZeroBasedCArray.fromList (map (*2) (elements :: [Float]))
      ===
      ZeroBasedCArray.amap (*2) (ZeroBasedCArray.fromList elements)
  ]

tests :: TestTree
tests = testGroup "ZeroBasedCArray"
        [ testsForType "Int" (Proxy :: Proxy Int)
        , testsForType "Complex Double" (Proxy :: Proxy (Complex Double))
        , amapTests
        ]


-- Helpers

wrapProxyInCArrayInt :: Proxy e -> Proxy (CArray Int e)
wrapProxyInCArrayInt = error "should not be called"

wrapProxyInList :: Proxy e -> Proxy [e]
wrapProxyInList = error "should not be called"

wrapProxyInZeroBasedCArray :: Proxy e -> Proxy (ZeroBasedCArray e)
wrapProxyInZeroBasedCArray = error "should not be called"
