module Main where

import Data.List (intersperse)
import qualified Data.Map as Map
import qualified Foreign.Storable as Storable
import qualified SoundStream.BitChord
import qualified SoundStream.Block as Block
import qualified SoundStream.CycleTimer
import SoundStream.Block ((!))
import qualified SoundStream.DFT as DFT
import qualified SoundStream.FourierHamming
import SoundStream.Frequency (fromHz, toHz)
import qualified SoundStream.HammingMelody
import qualified SoundStream.MultiCycleTimer
import qualified SoundStream.ShepardTimer
import qualified SoundStream.Sndio as Sndio
import SoundStream.Sndio (Handle, Parameters(Parameters), Sample)
import qualified SoundStream.TwistedHammingWeight
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)
import System.Environment (getArgs)
import qualified System.Exit as Exit
import System.IO (hPutBuf, stdout)

main :: IO ()
main =
  do args <- getArgs
     case parseArgs args of
       Nothing ->
         Exit.die
         ( "Usage: test_sounds [--mode (sndio|stdout)] " ++ soundNamesForUsage ++
           "\n--mode sndio (the default) plays the sound using the sndio library." ++
           "\n--mode stdout writes the sound to standard output as 16-bit signed samples, platform-native endianness, mono, 44.1kHz"
         )
       Just (mode, soundName) ->
         case Map.lookup soundName sounds of
           Nothing -> Exit.die ("Unknown sound: " ++ soundName)
           Just play ->
             case mode of
               SndioMode ->
                 do (handle, parameters) <- Sndio.open
                    play (Sndio.write handle) parameters
                    Sndio.close handle
               StdoutMode ->
                 let -- When updating this, be sure to update the usage message for --mode stdout too.
                     parameters = Parameters {Sndio.sampleRate = fromHz 44100}
                 in play writeBlockToStdout parameters

data Mode = SndioMode | StdoutMode

parseArgs :: [String] -> Maybe (Mode, String)
parseArgs ["--mode", "sndio", soundName] = Just (SndioMode, soundName)
parseArgs ["--mode", "stdout", soundName] = Just (StdoutMode, soundName)
parseArgs [soundName] = Just (SndioMode, soundName)
parseArgs _ = Nothing

sounds :: Map.Map String ((ZeroBasedCArray Sndio.Sample -> IO ()) -> Parameters -> IO ())
sounds = Map.fromList
 [ ("bit_chord", SoundStream.BitChord.play)
 , ("cycle_timer", SoundStream.CycleTimer.play)
 , ("hamming_weight", SoundStream.TwistedHammingWeight.play 0)
 , ("fh", SoundStream.FourierHamming.play)
 , ("hamming_melody", SoundStream.HammingMelody.play SoundStream.HammingMelody.HammingWeight)
 , ("multi_cycle_timer", SoundStream.MultiCycleTimer.play)
 , ("parity_hamming_melody", SoundStream.HammingMelody.play SoundStream.HammingMelody.HammingWeightAfterParity)
 , ("sine", playSine)
 , ("shepard_timer", SoundStream.ShepardTimer.play)
 , ("spectral_sawtooth", playSawtoothUsingDHT)
 , ("twisted_hamming_weight", SoundStream.TwistedHammingWeight.play (1/14))
 ]

-- The keys of sounds, formatted to be part of a usage message.
soundNamesForUsage :: String
soundNamesForUsage = concat ("(" : intersperse "|" (Map.keys sounds) ++ [")"])

writeBlockToStdout :: Block.Block b => b Sndio.Sample -> IO ()
writeBlockToStdout block =
  Block.withPtr block $
  \ ptr -> hPutBuf stdout ptr (Block.size block * Storable.sizeOf (block ! 0))

playSine :: (ZeroBasedCArray Sndio.Sample -> IO ()) -> Parameters -> IO ()
playSine write parameters =
  let sound = oneSecondOfSine parameters :: ZeroBasedCArray Sample
  in sequence_ (replicate 5 (write sound))

oneSecondOfSine :: Block.Block b => Parameters -> b Sample
oneSecondOfSine parameters =
  let numSamples = round (toHz (Sndio.sampleRate parameters))
      frequencyHz = 440
  in Block.fromList
     [ doubleToSample
       ( sin (2 * pi * frequencyHz * fromIntegral j / fromIntegral numSamples)
       )
     | j <- [0 .. numSamples - 1]
     ]

-- Play a sawtooth wave generated using the discrete Hartley transform. (The
-- phases probably aren't right so it's probably not actually a sawtooth wave.)
playSawtoothUsingDHT :: (ZeroBasedCArray Sndio.Sample -> IO ()) -> Parameters -> IO ()
playSawtoothUsingDHT write parameters =
  let -- Each block will be about 1/10-1/5 of a second long.
      blockSize = upToPowerOfTwo (toHz (Sndio.sampleRate parameters) / 10)
      blockDurationSeconds = fromIntegral blockSize / (toHz (Sndio.sampleRate parameters))
      fundamentalHz = 440
      fundamentalCyclesPerBlock = ceiling (fundamentalHz * blockDurationSeconds)
      spectrum = Block.fromList
                 [ if j > 0 && 2 * j <= blockSize && mod j fundamentalCyclesPerBlock == 0
                   then let harmonic = div j fundamentalCyclesPerBlock
                        in sqrt (fromIntegral blockSize) * (-1) ^ harmonic / fromIntegral harmonic
                   else 0
                 | j <- [0 .. blockSize - 1]
                 ]
      block = DFT.normalizedHartley spectrum :: ZeroBasedCArray Double
      sampleBlock = Block.bmap doubleToSample block
      durationSeconds = 5
  in sequence_
     $ replicate (ceiling (durationSeconds / blockDurationSeconds))
     $ write sampleBlock

upToPowerOfTwo :: (Fractional a, Ord a) => a -> Int
upToPowerOfTwo x | x < 1 = 1
                 | otherwise = 2 * (upToPowerOfTwo (x / 2))

doubleToSample :: Double -> Sample
doubleToSample x =
  let amplitude = fromIntegral (maxBound :: Sample) * 0.7
  in round (amplitude * x)
