module SoundStream.BitChord (play) where

import qualified Data.Array.IArray as IArray
import Data.Bits ((.&.))
import qualified SoundStream.Block as Block
import SoundStream.DFT (normalizedHartley)
import SoundStream.Duration (Duration, fromSeconds, toSeconds)
import SoundStream.Frequency (Frequency, fromHz, toHz)
import qualified SoundStream.Sndio as Sndio
import qualified SoundStream.ZeroBasedCArray as ZeroBasedCArray
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)

{-

  Parameters

-}

startTime :: Duration Double
startTime = fromSeconds (-30)
endTime = fromSeconds 0

tempo :: Frequency Double
tempo = fromHz 8

baseFrequency :: Frequency Double
baseFrequency = fromHz 110

{-

  Implementation

-}

play :: (ZeroBasedCArray Sndio.Sample -> IO ()) -> Sndio.Parameters -> IO ()
play write parameters =
  let blocks = soundAsBlockList parameters :: [ZeroBasedCArray Sndio.Sample]
  in sequence_ $ map write blocks

soundAsBlockList :: Sndio.Parameters -> [ZeroBasedCArray Sndio.Sample]
soundAsBlockList parameters =
  let sampleRateHz = toHz (Sndio.sampleRate parameters)
      blockSize = upToPowerOfTwo (sampleRateHz / toHz tempo / 10)
      blockDurationSec = fromIntegral blockSize / sampleRateHz
      baseFrequencyInCyclesPerBlock = ceiling (toHz baseFrequency * blockDurationSec)
  in [ blockForTime blockSize baseFrequencyInCyclesPerBlock (fromSeconds (toSeconds startTime + i * blockDurationSec))
     | i <- [0 .. (toSeconds endTime - toSeconds startTime) / blockDurationSec - 1]
     ]

upToPowerOfTwo :: Double -> Int
upToPowerOfTwo x
  | x < 1 = 1
  | otherwise = 2 * upToPowerOfTwo (x / 2)

blockForTime :: Int -> Int -> Duration Double -> ZeroBasedCArray Sndio.Sample
blockForTime blockSize baseFrequencyInCyclesPerBlock time =
  let beatNumber = floor (toSeconds time * toHz tempo) :: Int
      -- infiniteSparseSpectrum tests infinitely many bits of beatNumber.
      infiniteSparseSpectrum :: [(Int, Int)]
      infiniteSparseSpectrum = [ ( baseFrequencyInCyclesPerBlock * (i + 1)
                                 , if abs beatNumber .&. (2^i) == 0 then 0 else 1
                                 )
                               | i <- [0..]
                               ]
      sparseSpectrum :: [(Int, Int)]
      sparseSpectrum = takeWhile (\ (f, _) -> f * 2 < blockSize) infiniteSparseSpectrum
      spectrumSize = sum [1 / (fromIntegral f)^2 | f <- [1 .. length sparseSpectrum]]
      scale = sqrt (fromIntegral blockSize / (4 * spectrumSize))
      spectrum :: ZeroBasedCArray Double
      Just spectrum =
        ZeroBasedCArray.fromCArray $
        IArray.accumArray (+) 0 (0, blockSize - 1) $
        concat
        -- cas x - cas (-x) = 2 * sin x
        [ let v = fromIntegral x * scale / fromIntegral f
          in [(f - 1, v), (blockSize - f + 1, -v)]
        | (f, x) <- sparseSpectrum
        ]
  in Block.bmap doubleToSample $ normalizedHartley spectrum

doubleToSample :: Double -> Sndio.Sample
doubleToSample x =
  max minBound $ min maxBound $
  round
  ( fromIntegral (maxBound :: Sndio.Sample) * x
  )
