{-

Note: this sounds different depending on the sample rate.

Plays the Fourier transform of the some function, stretched out to cover the
duration of the timer.

I wrote this because I figured that since the timer sound is supposed to fit
exactly in a certain time range [start time, end time], that seems like a good
fit for defining the whole sound using one big Fourier transform. See comment
below about the function f.

-}

module SoundStream.FourierHamming (play) where

import qualified SoundStream.Block as Block
import SoundStream.Block (Block)
import SoundStream.Duration (Duration, fromSeconds, toSeconds)
import SoundStream.DFT (normalizedHartley)
import SoundStream.Frequency (Frequency, fromHz, toHz)
import qualified SoundStream.Sndio as Sndio
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)

{-

Parameters

-}

duration :: Duration Double
duration = fromSeconds 5

{-

Implementation

-}

play :: (ZeroBasedCArray Sndio.Sample -> IO ()) -> Sndio.Parameters -> IO ()
play write parameters =
  let renderedSound = sound parameters
  in do putStrLn "Rendering..."
        seq renderedSound $ putStrLn "Done."
        write renderedSound

-- We generate the sound as a single block, with one big Hartley transform.
sound :: Sndio.Parameters -> ZeroBasedCArray Sndio.Sample
sound parameters =
  let sampleRateHz = toHz (Sndio.sampleRate parameters)
      -- The block size will be a multiple of halfBlockSizeMultiple * 2.
      halfBlockSizeMultiple = 2^11
      halfBlockSize = halfBlockSizeMultiple *
                      ceiling (sampleRateHz * toSeconds duration / 2 / fromIntegral halfBlockSizeMultiple)
      halfSpectrum :: [Double]
      halfSpectrum = map spectrumStrengthAt [0 .. halfBlockSize - 1]
      {-
      Set up the phases so that every component starts and ends at zero (i.e.
      use sin rather than cos). Future work: try different phases. I just chose
      this way because it makes all the components start and end at zero and
      maybe that makes it less likely to click at the start and end.
      -}
      spectrum = halfSpectrum ++ map (* (-1)) (reverse halfSpectrum)
      unnormalizedOutput = normalizedHartley (Block.fromList spectrum) :: ZeroBasedCArray Double
  in normalize unnormalizedOutput

normalize :: Block b => b Double -> b Sndio.Sample
normalize x =
  let m = maximum $ Block.bmap abs $ Block.toList x
      sumSquare = sum $ map (^ 2) $ Block.toList x
      multiplyBy = 0.7 / m
  in Block.bmap (doubleToSample . (* multiplyBy)) x

doubleToSample :: Double -> Sndio.Sample
doubleToSample x
  | x >= 1 = error "too big"
  | x <= -1 = error "too small"
  | otherwise =
    round
    ( fromIntegral (maxBound :: Sndio.Sample) * x
    )

spectrumStrengthAt :: Int -> Double
spectrumStrengthAt x =
  let minFrequency = toSeconds duration * 20
      midFrequency = toSeconds duration * 60
      filterStrength = if fromIntegral x <= minFrequency
                       then 0
                       else min ((fromIntegral x - minFrequency) / (midFrequency - minFrequency)) (midFrequency / fromIntegral x)
  in if x == 0 then 0 else f x * filterStrength

{-

The sound is entirely defined by its spectrum, and the spectrum is the below
function (f) times a filter that cuts out especially high and low components.

f is defined so that numbers whose binary representation, read in order, have
many changes between 0 and 1 have higher strengths in the spectrum. In other
words, we emphasize numbers with lots of 1s if you use a gray code to write
them.

Why do it this way? The idea was that when a spectrum has a set of frequencies
bunched close together, you can hear that structure as beats, and maybe this
function will give lots of clumps with interesting local structure like that.
It doesn't seem to have resulted in a particularly interesting sound in
practice.

-}

f :: Int -> Double
f x = 7 ^ g x

g :: Int -> Int
g x | x == 0 = 0
    | mod (x+1) 4 >= 2 = 1 + g (div x 2)
    | otherwise = g (div x 2)
