{-

The Hamming weight function as a melody.

-}

module SoundStream.HammingMelody (Mode(..), play) where

import qualified SoundStream.Block as Block
import SoundStream.BlockUtil (groupListIntoBlocks)
import SoundStream.Duration (Duration, fromSeconds, toSeconds)
import SoundStream.Frequency (Frequency, fromHz, toHz)
import qualified SoundStream.Sndio as Sndio

{-

  Parameters

-}

-- The base waveform. Need only be defined on [0, 1). Output should be in (-1, 1).
baseWaveform :: Double -> Double
baseWaveform x = 2 * (if x <= 0.5 then x else 1 - x) - 0.5  -- triangle wave

startTime, endTime :: Duration Double
startTime = fromSeconds (-10)
endTime = fromSeconds 30

baseFrequency :: Frequency Double
baseFrequency = fromHz 110

data Mode = HammingWeight | HammingWeightAfterParity

{-

  Implementation

-}

play :: Block.Block b => Mode -> (b Sndio.Sample -> IO ()) -> Sndio.Parameters -> IO ()
play mode write parameters = playSampleList (soundAsSampleList mode parameters) write

playSampleList :: Block.Block b => [Sndio.Sample] -> (b Sndio.Sample -> IO ()) -> IO ()
playSampleList samples write =
  let blockSize = 1024
      blocks = groupListIntoBlocks blockSize samples
  in sequence_ $ map write blocks

soundAsSampleList :: Mode -> Sndio.Parameters -> [Sndio.Sample]
soundAsSampleList mode parameters =
  let sampleRateHz = toHz (Sndio.sampleRate parameters)
      secondsPerSample = 1 / sampleRateHz
      -- Phase should be in [0, 1).
      playFromPhaseAndTime :: Double -> Duration Double -> [Sndio.Sample]
      playFromPhaseAndTime phase time
        | time > endTime = []
        | otherwise =
          let nextPhase =
                nonIntegerPart
                (phase + toHz (frequencyAtTime mode time) / sampleRateHz)
          in doubleToSample (baseWaveform phase) :
             playFromPhaseAndTime nextPhase (fromSeconds (toSeconds time + secondsPerSample))
  in playFromPhaseAndTime 0 startTime

doubleToSample :: Double -> Sndio.Sample
doubleToSample x =
  max minBound $ min maxBound $
  round
  ( fromIntegral (maxBound :: Sndio.Sample) * 0.7 * x
  )

nonIntegerPart :: RealFrac a => a -> a
nonIntegerPart x = x - fromIntegral (floor x)

frequencyAtTime :: Mode -> Duration Double -> Frequency Double
frequencyAtTime mode time =
  let n = abs (floor (toSeconds time * 8))
      f = case mode of
            HammingWeight -> hammingWeight
            HammingWeightAfterParity -> hammingWeightAfterParity
  in fromHz (toHz baseFrequency * fromIntegral (f n))

-- Input should be non-negative.
hammingWeight :: Int -> Int
hammingWeight =
  let f acc x =
        if x == 0
        then acc
        else (f $! (acc + mod x 2)) (div x 2)
  in f 0

-- Input should be non-negative.
hammingWeightAfterParity :: Int -> Int
hammingWeightAfterParity =
  let f currentParity count x =
        if x == 0
        then (if currentParity == 1 then 0 else count)
        else let newParity = mod (currentParity + x) 2
             in ((f $! newParity) $! (count + newParity)) (div x 2)
  in f 0 0
