{-

Like CycleTimer, but has several cycles playing at the same time, with
parameters controlled by the top-level variable 'voices'.

-}

module SoundStream.MultiCycleTimer (play) where

import SoundStream.BlockUtil (groupListIntoBlocks)
import SoundStream.Duration
import SoundStream.Frequency
import qualified SoundStream.PulseTrack as PulseTrack
import SoundStream.PulseTrack (PulseTrack(PulseTrack))
import qualified SoundStream.Sndio as Sndio
import SoundStream.Sndio (Sample)
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)

{-

  Parameters

-}

-- The total length of the countdown, in seconds.
totalTimerDuration :: Int
totalTimerDuration = 32

-- For each voice, the number l of "lost notes", and the voice's pitch. When l
-- is zero, the voice simply plays one note per second. Higher values of l make
-- the voice lag more. More precisely: the voice playes a total of
-- (totalTimerDuration + 1 - l) notes over the whole interval of
-- totalTimerDuration seconds.
voices :: [(Int, Frequency Double)]
-- With this setting, we have one low pulse per second, and a slightly
-- higher-pitched cycle that takes the full totalTimerDuration to come back in
-- sync with the lowest note, then a higher pitch that takes half as long to
-- come back in sync (so comes back in sync twice, and finally one that comes
-- back in sync four times.
voices = [(0, fromHz 440), (1, fromHz 550), (2, fromHz 660), (4, fromHz 770)]

noteDuration :: Duration Double
noteDuration = fromSeconds 0.05

{-

  Implementation

-}

play :: (ZeroBasedCArray Sample -> IO ()) -> Sndio.Parameters -> IO ()
play write parameters =
  let tracks = map makeTrack voices
      soundStream = map doubleToSample (PulseTrack.render parameters tracks)
                    -- TODO: Figure out why the sound cuts off early if we
                    -- don't append some zeros like so:
                    ++ replicate 10000 0
      soundBlocks = groupListIntoBlocks 1024 soundStream
  in do sequence_ $ map write soundBlocks

makeTrack :: (Int, Frequency Double) -> PulseTrack
makeTrack (lostNotes, pitch) =
  let -- Since notes start sounding before the instant they're scheduled, we start the sound a little early.
      startTime = noteDuration
  in PulseTrack
     { PulseTrack.amplitude = 1 / fromIntegral (length voices)
     , PulseTrack.pitch = pitch
     , PulseTrack.duration = noteDuration
     , PulseTrack.centerTimes =
         [ fromSeconds
           ( toSeconds startTime
           + fromIntegral i
             * fromIntegral totalTimerDuration
             / fromIntegral (totalTimerDuration - lostNotes)
           )
         | i <- [0 .. totalTimerDuration - lostNotes]
         ]
     }

doubleToSample :: Double -> Sample
doubleToSample x =
  max minBound $ min maxBound $
  round
  ( fromIntegral (maxBound :: Sample) * x
  )
