module SoundStream.ShepardTimer (play) where

import Data.List (tails, transpose)
import Foreign.Storable (Storable)
import qualified SoundStream.Block as Block
import SoundStream.BlockUtil (groupListIntoBlocks)
import SoundStream.Block (Block)
import SoundStream.Duration (Duration, fromSeconds, toSeconds)
import SoundStream.Frequency (Frequency, fromHz, toHz)
import qualified SoundStream.Sndio as Sndio
import SoundStream.ZeroBasedCArray (ZeroBasedCArray)

{-

  Parameters

-}

{-
Frequencies between lowFrequency and highFrequency are included, with
midFrequency given the highest amplitude.
-}
lowFrequency, midFrequency, highFrequency :: Frequency Double
lowFrequency = fromHz 55
midFrequency = fromHz 220
highFrequency = fromHz 1760

-- Must be greater than 1.
interFrequencyRatio :: Double
interFrequencyRatio = 2

-- How quickly do they shift?

{-
When frequencyPower = 1, frequency scales linearly with time: e.g. the frequency
at 10 seconds will be 10 times the frequency at 1 second. In general, frequency
scales as |time| ** frequencyPower.
-}
frequencyPower :: Double
frequencyPower = 6

-- What range of time should we play?
startTime, endTime :: Duration Double
startTime = fromSeconds (-60)
endTime = fromSeconds 30

{-

  Implementation

-}

-- We set the amplitude to 1 / (max # tones that could be playing at once).
-- With some care, we could set this higher and still avoid clipping.
topLevelAmplitude :: Double
topLevelAmplitude = min 1 (1 / logBase interFrequencyRatio (toHz highFrequency / toHz lowFrequency))

play :: Block.Block b => (b Sndio.Sample -> IO ()) -> Sndio.Parameters -> IO ()
play = playFunction (startTime, endTime) soundAsFunction

playFunction :: Block.Block b =>
                (Duration Double, Duration Double) ->
                (Duration Double -> Double) -> (b Sndio.Sample -> IO ()) ->
                Sndio.Parameters -> IO ()
playFunction (startTime, endTime) f write parameters =
  let sampleRateHz = toHz (Sndio.sampleRate parameters)
      samples = [ let time = fromSeconds (fromIntegral sampleIndex / toHz (Sndio.sampleRate parameters))
                  in doubleToSample (f time)
                | sampleIndex <- [ round (toSeconds startTime * sampleRateHz)
                                   ..
                                   round (toSeconds endTime * sampleRateHz) - 1
                                 ]
                ]
      blockSize = 1024
      blocks = groupListIntoBlocks blockSize samples
  in sequence_ $ map write blocks

doubleToSample :: Double -> Sndio.Sample
doubleToSample x =
  max minBound $ min maxBound $
  round
  ( fromIntegral (maxBound :: Sndio.Sample) * x
  )

{-

The sound is a sum of infinitely many oscillators with changing frequency.

For j ranging over all integers, oscillator j's frequency at t seconds is
  freq j t = B * interFrequencyRatio ^^ j * abs t ** frequencyPower
where the constant B is chosen so that freq 0 0 = 1 Hz. To achieve this, we need
an argument to the sine function whose derivative with respect to t is freq j t.
We use:
  arg j t = B * interFrequencyRatio ^^ j * abs t ** (frequencyPower + 1) / (frequencyPower + 1)

-}
soundAsFunction :: Duration Double -> Double
soundAsFunction time =
  let timeSeconds = toSeconds time
      {-
      We don't want to compute an infinite sum, so we start by computing the
      range of j such that freq j t is between lowFrequency and highFrequency.
      -}
      lowestJ = ceiling
                ( logBase interFrequencyRatio (toHz lowFrequency) -
                  frequencyPower * logBase interFrequencyRatio (abs timeSeconds)
                )
      highestJ = floor
                 ( logBase interFrequencyRatio (toHz highFrequency) -
                   frequencyPower * logBase interFrequencyRatio (abs timeSeconds)
                 )
  in topLevelAmplitude *
     sum
     [ let baseFrequency = interFrequencyRatio ^^ j
           frequencyHz = baseFrequency * abs timeSeconds ** frequencyPower
           {-
           To determine the amplitude for each harmonic, we want a function
           that is zero at lowestFrequency and highestFrequency, and reaches 1
           at midFrequency.

           The below function has two parts: first it increases linearly from 0
           to 1, then it goes back down to 0 following the curve
           1 / ( f^2 - lowestFrequency^2 ) scaled so it starts at 1.
           -}
           amplitude = min
                       ((frequencyHz - toHz lowFrequency) / (toHz midFrequency - toHz lowFrequency))
                       ((1 / frequencyHz - 1 / toHz highFrequency)^2 / (1 / toHz midFrequency - 1 / toHz highFrequency)^2)
           {-
           Explained in the comment for soundAsFunction.
           -}
           arg = baseFrequency * abs timeSeconds ** (frequencyPower + 1) / (frequencyPower + 1)
       in amplitude * sin (2 * pi * arg)
     | j <- [lowestJ .. highestJ]
     ]
