{-

Note: this sounds different depending on the sample rate. This could probably
be fixed by transforming the sample number into a timestamp using the sample
rate instead of just using the sample number directly.

-}

module SoundStream.TwistedHammingWeight (play) where

import Data.Complex
import qualified SoundStream.Block as Block
import SoundStream.BlockUtil (groupListIntoBlocks)
import SoundStream.Duration (Duration, fromSeconds, toSeconds)
import SoundStream.Frequency (Frequency, fromHz, toHz)
import qualified SoundStream.Sndio as Sndio

{-

  Parameters

-}

startTime :: Duration Double
startTime = fromSeconds (-30)

{-

  Implementation

-}

-- zeroBitRotation is a parameter passed to the hammingWeightWithRotation
-- function.
play :: Block.Block b => Double -> (b Sndio.Sample -> IO ()) -> Sndio.Parameters -> IO ()
play zeroBitRotation write parameters = playSampleList (soundAsSampleList zeroBitRotation parameters) write

playSampleList :: Block.Block b => [Sndio.Sample] -> (b Sndio.Sample -> IO ()) -> IO ()
playSampleList samples write =
  let blockSize = 1024
      blocks = groupListIntoBlocks blockSize samples
  in sequence_ $ map write blocks

soundAsSampleList :: Double -> Sndio.Parameters -> [Sndio.Sample]
soundAsSampleList zeroBitRotation parameters =
  let startSampleIndex = round (toSeconds startTime * toHz (Sndio.sampleRate parameters))
  in
  [ doubleToSample (magnitude (twistedHammingWeight zeroBitRotation (abs n)) / 21 - 0.5)
  | n <- [startSampleIndex, startSampleIndex + 1 .. 0]
  ]

doubleToSample :: Double -> Sndio.Sample
doubleToSample x =
  max minBound $ min maxBound $
  round
  ( fromIntegral (maxBound :: Sndio.Sample) * 0.7 * x
  )

-- @twistedHammingWeight 0 n@ is the Hamming weight of @n@ (the number of bits
-- in its binary representation).
--
-- Passing a value other than 0 as the first parameter introduces a twist into
-- how the result is computed. It works like this. Start with 0, then consider
-- each bit of @n@ starting with the least significant. Every time you see a 1,
-- add 1. Every time you see a 0, rotate (in the complex plane) by an angle of
-- 2 * pi * zeroBitRotation. The effect should be that 1s grouped together
-- reinforce each other, but 1s separated by zeros result in pushes in
-- different directions, so don't reinferce each other as much (or interfere
-- with each other).
twistedHammingWeight :: Double -> Int -> Complex Double
twistedHammingWeight zeroBitRotation =
  let zeroMult = cis (2 * pi * zeroBitRotation)
      f acc x =
        if x == 0
        then acc
        else let acc' = if mod x 2 == 0 then acc * zeroMult else acc + 1
             in (f $! acc') (div x 2)
  in f 0
